@@ -36,21 +36,19 @@
 #include <sys/types.h>
 #include <sys/wait.h>
 #include <unistd.h>
-#include <string.h>
-#include <stdio.h>
+#include <inttypes.h>
 
 #include <glib.h>
 
-#include <openobex/obex.h>
-#include <openobex/obex_const.h>
-
+#include "obexd.h"
 #include "plugin.h"
 #include "log.h"
 #include "obex.h"
-#include "dbus.h"
+#include "manager.h"
 #include "mimetype.h"
 #include "service.h"
 #include "ftp.h"
+#include "filesystem.h"
 
 #define LST_TYPE "x-obex/folder-listing"
 #define CAP_TYPE "x-obex/capability"
@@ -83,7 +81,7 @@
     <sequence>								\
       <sequence>							\
         <uuid value=\"0x1106\"/>					\
-        <uint16 value=\"0x0100\" name=\"version\"/>			\
+        <uint16 value=\"0x0102\" name=\"version\"/>			\
       </sequence>							\
     </sequence>								\
   </attribute>								\
@@ -91,51 +89,8 @@
   <attribute id=\"0x0100\">						\
     <text value=\"%s\" name=\"name\"/>					\
   </attribute>								\
-</record>"
-
-#define PCSUITE_CHANNEL 24
-#define PCSUITE_WHO_SIZE 8
-
-#define PCSUITE_RECORD "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>	\
-<record>								\
-  <attribute id=\"0x0001\">						\
-    <sequence>								\
-      <uuid value=\"00005005-0000-1000-8000-0002ee000001\"/>		\
-    </sequence>								\
-  </attribute>								\
-									\
-  <attribute id=\"0x0004\">						\
-    <sequence>								\
-      <sequence>							\
-        <uuid value=\"0x0100\"/>					\
-      </sequence>							\
-      <sequence>							\
-        <uuid value=\"0x0003\"/>					\
-        <uint8 value=\"%u\" name=\"channel\"/>				\
-      </sequence>							\
-      <sequence>							\
-        <uuid value=\"0x0008\"/>					\
-      </sequence>							\
-    </sequence>								\
-  </attribute>								\
-									\
-  <attribute id=\"0x0005\">						\
-    <sequence>								\
-      <uuid value=\"0x1002\"/>						\
-    </sequence>								\
-  </attribute>								\
-									\
-  <attribute id=\"0x0009\">						\
-    <sequence>								\
-      <sequence>							\
-        <uuid value=\"00005005-0000-1000-8000-0002ee000001\"/>		\
-        <uint16 value=\"0x0100\" name=\"version\"/>			\
-      </sequence>							\
-    </sequence>								\
-  </attribute>								\
-									\
-  <attribute id=\"0x0100\">						\
-    <text value=\"%s\" name=\"name\"/>					\
+  <attribute id=\"0x0200\">						\
+    <uint16 value=\"%u\" name=\"psm\"/>					\
   </attribute>								\
 </record>"
 
@@ -143,20 +98,11 @@ static const uint8_t FTP_TARGET[TARGET_SIZE] = {
 			0xF9, 0xEC, 0x7B, 0xC4, 0x95, 0x3C, 0x11, 0xD2,
 			0x98, 0x4E, 0x52, 0x54, 0x00, 0xDC, 0x9E, 0x09 };
 
-static const uint8_t PCSUITE_WHO[PCSUITE_WHO_SIZE] = {
-			'P', 'C', ' ', 'S', 'u', 'i', 't', 'e' };
-
 struct ftp_session {
 	struct obex_session *os;
 	char *folder;
 };
 
-struct pcsuite_session {
-	struct ftp_session *ftp;
-	char *lock_file;
-	int fd;
-};
-
 static void set_folder(struct ftp_session *ftp, const char *new_folder)
 {
 	DBG("%p folder %s", ftp, new_folder);
@@ -169,7 +115,7 @@ static void set_folder(struct ftp_session *ftp, const char *new_folder)
 static int get_by_type(struct ftp_session *ftp, const char *type)
 {
 	struct obex_session *os = ftp->os;
-	const char *capability = obex_get_capability_path(os);
+	const char *capability = obex_option_capability();
 	const char *name = obex_get_name(os);
 	char *path;
 	int err;
@@ -179,9 +125,12 @@ static int get_by_type(struct ftp_session *ftp, const char *type)
 	if (type == NULL && name == NULL)
 		return -EBADR;
 
-	if (g_strcmp0(type, CAP_TYPE) == 0)
+	if (type != NULL && g_ascii_strcasecmp(type, CAP_TYPE) == 0)
 		return obex_get_stream_start(os, capability);
 
+	if (name != NULL && !is_filename(name))
+		return -EBADR;
+
 	path = g_build_filename(ftp->folder, name, NULL);
 	err = obex_get_stream_start(os, path);
 
@@ -197,7 +146,7 @@ void *ftp_connect(struct obex_session *os, int *err)
 
 	DBG("");
 
-	root_folder = obex_get_root_folder(os);
+	root_folder = obex_option_root_folder();
 
 	manager_register_session(os);
 
@@ -213,7 +162,7 @@ void *ftp_connect(struct obex_session *os, int *err)
 	return ftp;
 }
 
-int ftp_get(struct obex_session *os, obex_object_t *obj, void *user_data)
+int ftp_get(struct obex_session *os, void *user_data)
 {
 	struct ftp_session *ftp = user_data;
 	const char *type = obex_get_type(os);
@@ -263,6 +212,9 @@ int ftp_chkput(struct obex_session *os, void *user_data)
 	if (name == NULL)
 		return -EBADR;
 
+	if (!is_filename(name))
+		return -EBADR;
+
 	if (obex_get_size(os) == OBJECT_SIZE_DELETE)
 		return 0;
 
@@ -275,7 +227,7 @@ int ftp_chkput(struct obex_session *os, void *user_data)
 	return ret;
 }
 
-int ftp_put(struct obex_session *os, obex_object_t *obj, void *user_data)
+int ftp_put(struct obex_session *os, void *user_data)
 {
 	struct ftp_session *ftp = user_data;
 	const char *name = obex_get_name(os);
@@ -289,29 +241,32 @@ int ftp_put(struct obex_session *os, obex_object_t *obj, void *user_data)
 	if (name == NULL)
 		return -EBADR;
 
+	if (!is_filename(name))
+		return -EBADR;
+
 	if (size == OBJECT_SIZE_DELETE)
 		return ftp_delete(ftp, name);
 
 	return 0;
 }
 
-int ftp_setpath(struct obex_session *os, obex_object_t *obj, void *user_data)
+int ftp_setpath(struct obex_session *os, void *user_data)
 {
 	struct ftp_session *ftp = user_data;
 	const char *root_folder, *name;
-	uint8_t *nonhdr;
+	const uint8_t *nonhdr;
 	char *fullname;
 	struct stat dstat;
 	gboolean root;
 	int err;
 
-	if (OBEX_ObjectGetNonHdrData(obj, &nonhdr) != 2) {
+	if (obex_get_non_header_data(os, &nonhdr) != 2) {
 		error("Set path failed: flag and constants not found!");
 		return -EBADMSG;
 	}
 
 	name = obex_get_name(os);
-	root_folder = obex_get_root_folder(os);
+	root_folder = obex_option_root_folder();
 	root = g_str_equal(root_folder, ftp->folder);
 
 	DBG("%p name %s", ftp, name);
@@ -344,7 +299,7 @@ int ftp_setpath(struct obex_session *os, obex_object_t *obj, void *user_data)
 	}
 
 	/* Check and set to name path */
-	if (strstr(name, "/") || strcmp(name, "..") == 0) {
+	if (!is_filename(name)) {
 		error("Set path failed: name incorrect!");
 		return -EPERM;
 	}
@@ -353,10 +308,12 @@ int ftp_setpath(struct obex_session *os, obex_object_t *obj, void *user_data)
 
 	DBG("Fullname: %s", fullname);
 
-	if (root && obex_get_symlinks(os))
-		err = stat(fullname, &dstat);
-	else
-		err = lstat(fullname, &dstat);
+	err = verify_path(fullname);
+
+	if (err < 0)
+		goto done;
+
+	err = stat(fullname, &dstat);
 
 	if (err < 0) {
 		err = -errno;
@@ -364,8 +321,7 @@ int ftp_setpath(struct obex_session *os, obex_object_t *obj, void *user_data)
 		if (err == -ENOENT)
 			goto not_found;
 
-		DBG("%s: %s(%d)", root ? "stat" : "lstat",
-				strerror(-err), -err);
+		DBG("stat: %s(%d)", strerror(-err), -err);
 
 		goto done;
 	}
@@ -399,18 +355,47 @@ done:
 	return err;
 }
 
+static gboolean is_valid_path(const char *path)
+{
+	gchar **elements, **cur;
+	int depth = 0;
+
+	elements = g_strsplit(path, "/", 0);
+
+	for (cur = elements; *cur != NULL; cur++) {
+		if (**cur == '\0' || strcmp(*cur, ".") == 0)
+			continue;
+
+		if (strcmp(*cur, "..") == 0) {
+			depth--;
+			if (depth < 0)
+				break;
+			continue;
+		}
+
+		depth++;
+	}
+
+	g_strfreev(elements);
+
+	if (depth < 0)
+		return FALSE;
+
+	return TRUE;
+}
+
 static char *ftp_build_filename(struct ftp_session *ftp, const char *destname)
 {
 	char *filename;
 
 	/* DestName can either be relative or absolute (FTP style) */
-	if (g_path_is_absolute(destname))
-		filename = g_build_filename(destname, NULL);
+	if (destname[0] == '/')
+		filename = g_build_filename(obex_option_root_folder(),
+                                                                destname, NULL);
 	else
 		filename = g_build_filename(ftp->folder, destname, NULL);
 
-	/* Check if destination is inside root path */
-	if (g_str_has_prefix(filename, ftp->folder))
+	if (is_valid_path(filename + strlen(obex_option_root_folder())))
 		return filename;
 
 	g_free(filename);
@@ -421,7 +406,7 @@ static char *ftp_build_filename(struct ftp_session *ftp, const char *destname)
 static int ftp_copy(struct ftp_session *ftp, const char *name,
 							const char *destname)
 {
-	char *source, *destination;
+	char *source, *destination, *destdir;
 	int ret;
 
 	DBG("%p name %s destination %s", ftp, name, destname);
@@ -436,6 +421,16 @@ static int ftp_copy(struct ftp_session *ftp, const char *name,
 
 	destination = ftp_build_filename(ftp, destname);
 
+	if (destination == NULL)
+		return -EBADR;
+
+	destdir = g_path_get_dirname(destination);
+	ret = verify_path(destdir);
+	g_free(destdir);
+
+	if (ret < 0)
+		return ret;
+
 	source = g_build_filename(ftp->folder, name, NULL);
 
 	ret = obex_copy(ftp->os, source, destination);
@@ -449,7 +444,7 @@ static int ftp_copy(struct ftp_session *ftp, const char *name,
 static int ftp_move(struct ftp_session *ftp, const char *name,
 							const char *destname)
 {
-	char *source, *destination;
+	char *source, *destination, *destdir;
 	int ret;
 
 	DBG("%p name %s destname %s", ftp, name, destname);
@@ -464,6 +459,16 @@ static int ftp_move(struct ftp_session *ftp, const char *name,
 
 	destination = ftp_build_filename(ftp, destname);
 
+	if (destination == NULL)
+		return -EBADR;
+
+	destdir = g_path_get_dirname(destination);
+	ret = verify_path(destdir);
+	g_free(destdir);
+
+	if (ret < 0)
+		return ret;
+
 	source = g_build_filename(ftp->folder, name, NULL);
 
 	ret = obex_move(ftp->os, source, destination);
@@ -474,13 +479,16 @@ static int ftp_move(struct ftp_session *ftp, const char *name,
 	return ret;
 }
 
-int ftp_action(struct obex_session *os, obex_object_t *obj, void *user_data)
+int ftp_action(struct obex_session *os, void *user_data)
 {
 	struct ftp_session *ftp = user_data;
 	const char *name, *destname;
 	uint8_t action_id;
 
 	name = obex_get_name(os);
+	if (name == NULL || !is_filename(name))
+		return -EBADR;
+
 	destname = obex_get_destname(os);
 	action_id = obex_get_action_id(os);
 
@@ -492,7 +500,7 @@ int ftp_action(struct obex_session *os, obex_object_t *obj, void *user_data)
 	case 0x01: /* Move/Rename Object */
 		return ftp_move(ftp, name, destname);
 	default:
-		return -EINVAL;
+		return -ENOSYS;
 	}
 }
 
@@ -512,6 +520,8 @@ static struct obex_service_driver ftp = {
 	.name = "File Transfer server",
 	.service = OBEX_FTP,
 	.channel = FTP_CHANNEL,
+	.port = OBEX_PORT_RANDOM,
+	.secure = TRUE,
 	.record = FTP_RECORD,
 	.target = FTP_TARGET,
 	.target_size = TARGET_SIZE,
